signature UNIVERSUM =  
  sig 
      type obj 
      val objects : obj list 
  end        

functor Quantoren (U : UNIVERSUM) =
  struct 
    type obj = U.obj
    val objects = U.objects

    exception Undefined

    (* einstellige verallgemeinerte Quantoren *) 
    type Quant = (obj -> bool) -> (obj -> bool) -> bool

    fun thefirst(A) = case (List.find A objects) 
                       of SOME x => x 
                        | NONE => raise Undefined

    fun forall A G = List.all G (List.filter A objects)
    fun exists A G = List.exists G (List.filter A objects)

    fun most A G = 
        let 
            val As =  List.filter A objects  
            val GAs = List.filter G As       
            val nGAs = List.filter (not o G) As
        in 
            (List.length GAs) > (List.length nGAs)
        end

    (* zweistellige verallgemeinerte Quantoren *)
    type QQuant = (obj -> bool)*(obj -> bool) -> (obj -> bool) -> bool

    fun more (A,B) G = 
        let 
            val As = List.filter A objects
            val Bs = List.filter B objects
            val GAs = List.filter G As
            val GBs = List.filter G Bs
        in 
            List.length GAs > List.length GBs
        end

    (* verallgemeinerte Quantoren fuer Paare von Objekten *)

    val object_pairs = 
        List.foldr (op @) [] 
                   (List.map (fn x => (List.map (fn y => (x,y)) objects)) objects)

    fun forall_2_2 A G = List.all G (List.filter A object_pairs)
    fun exists_2_2 A G = List.exists G (List.filter A object_pairs)
    fun forall_1_2 A G = forall A (fn x => forall A (fn y => G(x,y)))  (* ggf.: x=y or G(x,y) *)
    fun exists_1_2 A G = exists A (fn x => exists A (fn y => G(x,y)))  (* ggf.: x=/=y and G(x,y) *)
  end

